#include "Heap.h"

extern PageDirectory *KernelDir;			/* Kernel's PageDirectory. */

Heap *TheHeap = 0;							/* The actual heap. */

/* Searches for the smallest hole that will fit the specified size. */
static int SeekHole(unsigned Size, bool PageAlign, Heap *_Heap)
{
	/* Declare a few variables to prevent variable recreation inside the loop. */
	unsigned i = 0;
	unsigned Location = 0;

	int Offset = 0;
	int HoleSize = 0;

	HeapHeader *Header = 0x0;

	/* Parse the index. */
	while(i < _Heap->Index.Size)
	{
		Header = reinterpret_cast<HeapHeader *>(RetrieveListItem(&_Heap->Index, i));
		
		if(PageAlign)
		{
			Location = (unsigned) (Header);

			Offset = 0;

			if( ((Location + sizeof(HeapHeader)) & 0xFFFFF000) )
				Offset = PAGE_SIZE - (Location + sizeof(HeapHeader)) % PAGE_SIZE;

			HoleSize = static_cast<int>(Header->Size - Offset);

			/* Check if the size of the hole is big enough to fit the requested size. */
			if(HoleSize >= static_cast<int>(Size))
				break;
		}

		else if(Header->Size >= Size)
			break;

		i++;
	}

	/* Nothing found, i is at the end of the index. */
	if(i == _Heap->Index.Size)
		return -1;
	
	/* We found some space, return the location. */
	return i;
}

/* Expands the heap to the specified NewSize. */
static bool ExpandHeap(unsigned NewSize, Heap *_Heap)
{
	/*
	 * Can't expand the heap to a smaller size than it is right now.
	 * Expanding the heap to the same size it is currently is useless.
	 * These aren't expanding!
	 */
	if(NewSize <= (_Heap->EndAddress - _Heap->StartAddress))
		return false;

	/* Page-align if needed. */
	if(NewSize & 0xFFFFF000 != 0)
	{
		NewSize &= 0xFFFFF000;
		NewSize += PAGE_SIZE;
	}
	
	/* Check to see if expanding will move beyond the MaxAddress. */
	if((_Heap->StartAddress + NewSize) > _Heap->MaxAddress)
		return false;

	unsigned OldSize = (_Heap->EndAddress - _Heap->StartAddress);
	unsigned i = OldSize;

	/* Start from the end of the heap, parsing the old contents is useless anyway. */
	while(i < NewSize)
	{
		AllocFrame(GetPage(_Heap->StartAddress + i, 1, KernelDir), _Heap->SupervisorOnlyNewPages? true : false,
			_Heap->ReadOnlyNewPages? false : true);

		i += PAGE_SIZE;
	}

	/* Set the size to its new value. */
	_Heap->EndAddress = (_Heap->StartAddress + NewSize);

	return true;
}

/* Contracts the heap to the specified NewSize, making it smaller and freeing unnecessary frames. */
static int ContractHeap(unsigned NewSize, Heap *_Heap)
{
   	if(NewSize >= (_Heap->EndAddress - _Heap->StartAddress))
		return -1;

	if(NewSize & PAGE_SIZE)
	{
		NewSize &= PAGE_SIZE;
		NewSize += PAGE_SIZE;
	}

	if(NewSize < HEAP_MINIMUM_SIZE)
		NewSize = HEAP_MINIMUM_SIZE;

	unsigned OldSize = (_Heap->EndAddress - _Heap->StartAddress);
	unsigned i = (OldSize - PAGE_SIZE);

	while(NewSize < i)
	{
		FreeFrame(GetPage(_Heap->StartAddress + i, 0, KernelDir));
		i -= PAGE_SIZE;
	}

	_Heap->EndAddress = (_Heap->StartAddress + NewSize);

	return NewSize;
}

static bool HeapHeaderIsLessThan(void *a, void *b)
{
	return ((reinterpret_cast<HeapHeader *>(a)->Size < reinterpret_cast<HeapHeader *>(b)->Size)? true : false);
}

Heap *CreateHeap(unsigned StartAddress, unsigned EndAddress, unsigned MaxAddress, bool SupervisorOnly, bool ReadOnly)
{
	Heap *NewHeap = reinterpret_cast<Heap *>(Memory->Alloc(sizeof(Heap)));

	/* Check if StartAddress and EndAddress are page-aligned. */
	if( (!(StartAddress % PAGE_SIZE == 0)) || (!(EndAddress % PAGE_SIZE == 0)) )
		return 0x0;

	/* Create the index. */
	NewHeap->Index = PutOrderedList((void *) (StartAddress), HEAP_INDEX_SIZE, &HeapHeaderIsLessThan);
	StartAddress += (sizeof(void *) * HEAP_INDEX_SIZE);

	/* Check if StartAddress is page-aligned, if it isn't, page-align it. */
	if(StartAddress & 0xFFFFF000 != 0)
	{
		StartAddress &= 0xFFFFF000;
		StartAddress += PAGE_SIZE;
	}

	/* Update the new heap object. */
	NewHeap->StartAddress = StartAddress;
	NewHeap->EndAddress = EndAddress;
	NewHeap->MaxAddress = MaxAddress;
	NewHeap->SupervisorOnlyNewPages = SupervisorOnly;
	NewHeap->ReadOnlyNewPages = ReadOnly;

	/* The index is assumed to be empty when creating the heap, so it must be seen as a hole. */
	HeapHeader *HugeHole = reinterpret_cast<HeapHeader *>(StartAddress);
	HugeHole->Size = (EndAddress - StartAddress);
	HugeHole->Magic = HEAP_MAGIC_NUMBER;
	HugeHole->IsHole = true;
	
	/* Dump HugeHole into the NewHeap's index. */
	InsertListItem(&NewHeap->Index, HugeHole);

	/* And done! */
	return NewHeap;
}

void *HeapAlloc(unsigned Size, bool PageAlign, Heap *_Heap)
{
	/* Make Size header and footer inclusive and find a fitting hole. */
	unsigned RealSize = (Size + sizeof(HeapHeader) + sizeof(HeapFooter));
	int i = SeekHole(RealSize, PageAlign, _Heap);

	/* Was a hole found? */
	if(i == -1)
	{
		/* Store old data. */
		unsigned TmpLength = (_Heap->EndAddress - _Heap->StartAddress);
		unsigned TmpEndAddress = _Heap->EndAddress;

		/* Try to allocate more space; */
		if(!ExpandHeap(TmpLength + RealSize, _Heap))
			return 0x0;

		unsigned NewLength = (_Heap->EndAddress - _Heap->StartAddress);
		i = 0;

		unsigned Index = static_cast<unsigned>(-1);
		unsigned Value = 0;
		unsigned TmpVar = 0;

		while(i < static_cast<int>(_Heap->Index.Size))
		{
			TmpVar = (unsigned) (RetrieveListItem(&_Heap->Index, i));

			if(TmpVar > Value)
			{
				Value = TmpVar;
				Index = i;
			}

			++i;
		}

		/* No headers found? */
		if(Index == static_cast<unsigned>(-1))
		{
			HeapHeader *Header = (HeapHeader *) (TmpEndAddress);
			Header->Magic = HEAP_MAGIC_NUMBER;
			Header->Size = NewLength - TmpLength;
			Header->IsHole = true;

			HeapFooter *Footer = (HeapFooter *) (TmpEndAddress + Header->Size - sizeof(HeapFooter));
			Footer->Magic = HEAP_MAGIC_NUMBER;
			Footer->Header = Header;
			
			InsertListItem(&_Heap->Index, Header);
		}

		else
		{
			/* Update the last header. */
			HeapHeader *Header = (HeapHeader *) RetrieveListItem(&_Heap->Index, Index);
			Header->Size += NewLength - TmpLength;

			/* Update the accompanying footer. */
			HeapFooter *Footer = (HeapFooter *) ((unsigned) Header + Header->Size - sizeof(HeapFooter));
			Footer->Magic = HEAP_MAGIC_NUMBER;
			Footer->Header = Header;
		}

		/* Memory created, try again. */
		return HeapAlloc(Size, PageAlign, _Heap);
	}

	/* Store more old values. */
	HeapHeader *OldHoleHeader = reinterpret_cast<HeapHeader *>(RetrieveListItem(&_Heap->Index, i));

	unsigned OldHolePos = (unsigned) OldHoleHeader;
	unsigned OldHoleSize = OldHoleHeader->Size;

	if(OldHoleSize - RealSize < sizeof(HeapHeader) + sizeof(HeapFooter))
	{
		Size += OldHoleSize - RealSize;
		RealSize = OldHoleSize;
	}
	
	/* Needs page-align? */
	if(PageAlign && OldHolePos & 0xFFFFF000)
	{
		unsigned NewLoc = (OldHolePos + PAGE_SIZE - (OldHolePos & 0xFFF) - sizeof(HeapHeader));

		HeapHeader *HoleHeader = (HeapHeader *) OldHoleHeader;

		HoleHeader->Size = (PAGE_SIZE - (OldHolePos & 0xFFF) - sizeof(HeapHeader));
		HoleHeader->Magic = HEAP_MAGIC_NUMBER;
		HoleHeader->IsHole = true;

		HeapFooter *HoleFooter = (HeapFooter *) ((unsigned) NewLoc - sizeof(HeapFooter));

		HoleFooter->Magic = HEAP_MAGIC_NUMBER;
		HoleFooter->Header = HoleHeader;

		OldHolePos = NewLoc;
		OldHoleSize -= HoleHeader->Size;
	}
	
	/* This hole is useless. */
	else
		DeleteListItem(&_Heap->Index, i);

	/* Overwrite original header. */
	HeapHeader *BlockHeader = reinterpret_cast<HeapHeader *>(OldHolePos);

	BlockHeader->Magic = HEAP_MAGIC_NUMBER;
	BlockHeader->IsHole = false;
	BlockHeader->Size = RealSize;

	/* Overwrite original footer. */
	HeapFooter *BlockFooter = reinterpret_cast<HeapFooter *>(OldHolePos + sizeof(HeapHeader) + Size);

	BlockFooter->Magic = HEAP_MAGIC_NUMBER;
	BlockFooter->Header = BlockHeader;

	/* New hole write needed after allocated block? */
	if(OldHoleSize - RealSize > 0)
	{
		HeapHeader *HoleHeader = reinterpret_cast<HeapHeader *>(OldHolePos + sizeof(HeapHeader) + Size + sizeof(HeapFooter));

		HoleHeader->Magic = HEAP_MAGIC_NUMBER;
		HoleHeader->IsHole = true;
		HoleHeader->Size = OldHoleSize - RealSize;

		HeapFooter *HoleFooter = reinterpret_cast<HeapFooter *>((unsigned) HoleHeader + OldHoleSize - RealSize - sizeof(HeapFooter));

		if((unsigned) HoleFooter < _Heap->EndAddress)
		{
			HoleFooter->Magic = HEAP_MAGIC_NUMBER; //<- Fuck up
			HoleFooter->Header = HoleHeader;
		}

		/* Place the new hole in the heap index. */
		InsertListItem(&_Heap->Index, HoleHeader);
	}

	/* Done. */
	return (reinterpret_cast<void *>((unsigned) BlockHeader + sizeof(HeapHeader)));
}

void HeapFree(void *Memory, Heap *_Heap)
{
	if(!Memory)
		return;

	/* Grab header and footer of the specified memory. */
	HeapHeader *Header = (HeapHeader *) ((unsigned) Memory - sizeof(HeapHeader));
	HeapFooter *Footer = (HeapFooter *) ((unsigned) Header + Header->Size - sizeof(HeapFooter));

	if(Header->Magic != HEAP_MAGIC_NUMBER || Footer->Magic != HEAP_MAGIC_NUMBER)
		return;

	/* Clear garbage memory. */
	::Memory->Set(Memory, 0, ((unsigned) Memory < (unsigned) Footer));

	/* This header now defines a hole. */
	Header->IsHole = true;
	bool AddHeaderToIndex = true;

	/* Try to unify this header with the footer left of this header. */
	HeapFooter *LeftFooter = reinterpret_cast<HeapFooter *>((unsigned) Header - sizeof(HeapFooter));

	/* Is this footer valid and does it define a hole as well? */
	if(LeftFooter->Magic == HEAP_MAGIC_NUMBER && LeftFooter->Header->IsHole)
	{
		/*
		 * The current header and left footer are removed and the header going with
		 * the left footer is the header of the new unified block. The current footer
		 * remains the footer of the unified block.
		 */
		unsigned CacheSize = Header->Size;
		Header = LeftFooter->Header;
		Footer->Header = Header;
		Header->Size += CacheSize;

		/* This header is already in the index. */
		AddHeaderToIndex = false;
	}

	/* Try to unify this footer with the header right of this footer. */
	HeapHeader *RightHeader = reinterpret_cast<HeapHeader *>((unsigned) Footer + sizeof(HeapFooter));

	/* Is this header valid and does it define a hole as well? */
	if(RightHeader->Magic == HEAP_MAGIC_NUMBER && RightHeader->IsHole)
	{
		/*
		 * The current header is the header of the unified block and the
		 * footer of the unified block is the footer that goes with the
		 * header right of the current block.
		 */
		Header->Size += RightHeader->Size;
		LeftFooter = (HeapFooter *) ((unsigned) RightHeader + RightHeader->Size - sizeof(HeapFooter));

		Footer = LeftFooter;

		unsigned i = 0;

		/* Remove the RightHeader from the index as it is now part of a unified block. */
		while((i < _Heap->Index.Size) && (RetrieveListItem(&_Heap->Index, i) != (void *) RightHeader))
			++i;

		/* Is i valid? */
		if(i >= _Heap->Index.Size)
			return;

		/* Remove the entry. */
		DeleteListItem(&_Heap->Index, i);
	}

	/* Is contraction possible? */
	if((unsigned) Footer + sizeof(HeapFooter) == _Heap->EndAddress)
	{
		unsigned OldLength = (_Heap->EndAddress - _Heap->StartAddress);
		unsigned NewLength = ContractHeap((unsigned) Header - _Heap->StartAddress, _Heap);

		/* Does the block still exist? */
		if(Header->Size - (OldLength - NewLength) > 0)
		{
			Header->Size -= OldLength - NewLength;

			Footer = (HeapFooter *) ((unsigned) Header + Header->Size - sizeof(HeapFooter));
			Footer->Magic = HEAP_MAGIC_NUMBER;
			Footer->Header = Header;
		}

		else
		{
			unsigned i = 0;

			while((i < _Heap->Index.Size) && (RetrieveListItem(&_Heap->Index, i) != reinterpret_cast<void *>(RightHeader)))
				++i;

			if(i < _Heap->Index.Size)
				DeleteListItem(&_Heap->Index, i);
		}
	}

	if(AddHeaderToIndex)
		InsertListItem(&_Heap->Index, Header);
}
